
import sys
import os
from PyQt5 import QtWidgets, QtGui, QtCore

import mathLib as cm

from PyQt5.QtCore import Qt, QUrl
from PyQt5.QtMultimedia import QMediaPlayer, QMediaContent
from PyQt5.QtGui import QMovie
from PyQt5.QtWidgets import QLabel

class CalculatorApp(QtWidgets.QWidget):
    ## Constructor for CalculatorApp.
    #  Initializes the application's UI and main settings.
    def __init__(self):
        super().__init__()
        self.initUI()
    
    ## Initializes the User Interface for the calculator.
    #  Sets window geometry, title, background image, and widgets.
    def initUI(self):
        self.setGeometry(300, 300, 350, 600)  # x, y, width, height
        self.setWindowTitle('Custom Math Calculator')
        
        # Load and set the background image
        self.setAutoFillBackground(True)
        p = self.palette()
        p.setColor(self.backgroundRole(), QtGui.QColor('white'))  # Fallback background color
        bg_image = QtGui.QPixmap('background.png').scaled(self.size(), aspectRatioMode=QtCore.Qt.IgnoreAspectRatio, transformMode=QtCore.Qt.SmoothTransformation)
        p.setBrush(self.backgroundRole(), QtGui.QBrush(bg_image))
        self.setPalette(p)
        
        # Initialize the display for calculations
        self.current_input = ""
        self.result = None
        self.operation = None
        
        # Layout and widgets
        self.create_widgets()
        
    ## Handles resizing of the application window.
    #  @param event The event object containing details of the resize event.
    def resizeEvent(self, event):
        bg_image = QtGui.QPixmap('background.png').scaled(self.size(), aspectRatioMode=QtCore.Qt.IgnoreAspectRatio, transformMode=QtCore.Qt.SmoothTransformation)
        p = self.palette()
        p.setBrush(self.backgroundRole(), QtGui.QBrush(bg_image))
        self.setPalette(p)
        super().resizeEvent(event)
        
    ## Creates and arranges the widgets in the calculator.
    def create_widgets(self):
        grid = QtWidgets.QGridLayout()
        self.setLayout(grid)

        # Create and configure the display label
        self.display_label = QtWidgets.QLabel('', self)
        self.display_label.setStyleSheet("QLabel { background-color : rgba(255, 255, 255, 75); font-size: 35pt; color: black; text-align: right; padding-top: 50px; padding-bottom: 50px; padding-left: 0px; padding-right: 0px; }")
        self.display_label.setFont(QtGui.QFont("DejaVu Sans", 24))
        grid.addWidget(self.display_label, 0, 0, 1, 4, alignment=QtCore.Qt.AlignBottom)

        # Define button labels and positions
        btn_texts = [
            ('C', 2), ('sqrt', 2), ('pow', 2), ('/', 2),
            ('7', 3), ('8', 3), ('9', 3), ('*', 3),
            ('4', 4), ('5', 4), ('6', 4), ('-', 4),
            ('1', 5), ('2', 5), ('3', 5), ('+', 5),
            ('More...', 6), ('0', 6), ('neg', 6), ('=', 6), 
        ]

        # Initialize buttons and connect signals
        for text, row in btn_texts:
            col = btn_texts.index((text, row)) % 4
            button = QtWidgets.QPushButton(text, self)
            button.setStyleSheet("QPushButton { font-size: 14pt; width: 60px; height: 60px; background-color: #CCCCCC; }")
            if text == 'More...':
                menu = self.create_special_operations_menu(button)
                button.setMenu(menu)
            else:
                # Ensure lambda captures 'text' value at the time of definition, not at execution time
                button.clicked.connect(lambda checked, b=text: self.on_button_press(b))
            grid.addWidget(button, row, col)


        grid.setColumnStretch(1, 4)
        grid.setRowStretch(1, 5)

    ## Creates a menu for special operations like square root, factorial, etc.
    #  @param parent The parent widget that the menu will be attached to.
    #  @return The constructed QMenu object.
    def create_special_operations_menu(self, parent):
        menu = QtWidgets.QMenu(parent)
        for action_text in ['fact', 'abs', 'ln']:
            action = QtWidgets.QAction(action_text, parent)
            action.triggered.connect(lambda checked, a=action_text: self.perform_special_operation(a))
            menu.addAction(action)
        return menu
    
    ## Displays an animation and sound effect when division by zero occurs.
    def show_division_by_zero_animation(self):
        # Setup paths for animation and sound
        script_dir = os.path.dirname(os.path.realpath(__file__))
        gif_path = os.path.join(script_dir, "divide_by_zero.gif")
        sound_file_path = os.path.join(script_dir, "error_sound.mp3")

        # Ensure the label for the animation exists
        if hasattr(self, 'animation_label') and self.animation_label:
            self.animation_label.movie().start()
            self.animation_label.show()
            self.animation_label.raise_()
        else:
            self.animation_label = QLabel(self)
            self.animation_label.setGeometry(50, 50, 250, 250)  # Adjust size and position as needed
            movie = QMovie(gif_path)
            if movie.isValid():
                self.animation_label.setMovie(movie)
                movie.start()
                self.animation_label.show()
                self.animation_label.raise_()
            else:
                print("Failed to load the GIF")  # Debug print

        # Setup the media player for sound effects
        if not hasattr(self, 'media_player'):
            self.media_player = QMediaPlayer()
            self.media_player.mediaStatusChanged.connect(self.handleMediaStatusChanged)
        self.media_player.setMedia(QMediaContent(QUrl.fromLocalFile(sound_file_path)))
        self.media_player.play()

    ## Replays sound effect in a loop as needed.
    #  @param status The current media status.
    def handleMediaStatusChanged(self, status):
        if status == QMediaPlayer.EndOfMedia:
            self.media_player.play()  # Replay the sound for looping effect

    ## Processes button presses, handles operation logic, and updates the display.
    #  @param value The value or operation from the button press.
    def on_button_press(self, value):
        # Handle button press for clear, negation, operations, equals, and numerical input
        if value == 'C':
            self.clear()
        elif value == 'neg':
            # Negate the current input
            if self.current_input:
                self.current_input = str(-float(self.current_input))
                self.display_label.setText(self.display_label.text()[:-len(self.current_input)] + self.current_input)
            elif self.result is not None:
                # Negate the result if no current input is available
                self.result = -self.result
                self.display_label.setText(str(self.result))
        elif value in {'+', '-', '*', '/', 'pow', 'sqrt'}:
            # Check if operation can be inputted (not first or without any preceding numbers)
            if self.current_input == '' and (self.result is None or self.display_label.text() == ''):
                return  # Prevent operation as first input
            # Store current operation and reset current input
            self.operation = value
            # Only capture the current result if input has been entered
            self.result = float(self.current_input) if self.current_input else self.result
            self.current_input = ""
            # Display the operation symbol on the display label
            if value == 'pow':
                self.display_label.setText(self.display_label.text() + " ^ ")
            elif value == 'sqrt':
                self.display_label.setText(self.display_label.text() + " √ ")
            else:
                self.display_label.setText(self.display_label.text() + " " + value + " ")
        elif value == '=':
            # Execute the calculation if there is an operation to perform
            if self.current_input and self.operation:
                self.calculate_result()
                self.operation = None
                self.current_input = ""
        else:
            # Append the number or decimal point to current input and update display
            self.current_input += value
            self.display_label.setText(self.display_label.text() + value)


    ## Calculates the result based on the current operation and operand.
    def calculate_result(self):
        # Handle operations and check for errors
        try:
            second_operand = float(self.current_input)
            if self.operation == '+':
                self.result = cm.add(self.result, second_operand)
            elif self.operation == '-':
                self.result = cm.sub(self.result, second_operand)
            elif self.operation == '*':
                self.result = cm.mul(self.result, second_operand)
            elif self.operation == '/':
                if second_operand == 0:
                    raise ValueError("Cannot divide by zero.")
                self.result = cm.div(self.result, second_operand)
            elif self.operation == 'pow':
                self.result = cm.powr(self.result, second_operand)
            elif self.operation == 'sqrt':
                self.result = cm.sqrt(self.result, second_operand)

        except ValueError as e:
            if str(e) == "Cannot divide by zero.":
                self.show_division_by_zero_animation()  # Call the animation for division by zero
            else:
                self.display_label.setText("Error: " + str(e))  # Handle other errors normally
            return  # Exit the method early
        
        # If no exceptions, update display with result
        self.display_label.setText(str(self.result))

    ## Performs a specified special mathematical operation.
    #  @param operation The mathematical operation to perform.
    def perform_special_operation(self, operation):
        # Handle special operations
        if not self.current_input:
            return  # Exit if there's no input
        try:
            operand = float(self.current_input)
            if operation == 'sqrt':
                self.result = cm.sqrt(operand)
            elif operation == 'ln':
                self.result = cm.log(operand)
            elif operation == 'abs':
                self.result = cm.abs(operand)
            elif operation == 'fact':
                self.result = cm.fact(int(operand))
            # Update display with result
            self.display_label.setText(str(self.result))
        except ValueError as e:
            self.display_label.setText("Error: " + str(e))

    ## Clears the calculator's display and state.
    def clear(self):
        # Reset calculator state and clear display
        self.current_input = ""
        self.result = None
        self.operation = None
        self.display_label.setText("")

        # Stop the animation and sound effects
        if hasattr(self, 'animation_label') and self.animation_label.movie():
            self.animation_label.movie().stop()
            self.animation_label.hide()
        if hasattr(self, 'media_player'):
            self.media_player.stop()

           # Capture key presses
    def keyPressEvent(self, event):
        key = event.key()
        if key == Qt.Key_Enter or key == Qt.Key_Return:
            self.on_button_press('=')
        elif key == Qt.Key_Backspace:
            if self.current_input:
                self.current_input = self.current_input[:-1]
                self.display_label.setText(self.display_label.text()[:-1])
        elif key == Qt.Key_Escape:
            self.clear()
        elif key in range(Qt.Key_0, Qt.Key_9 + 1):
            self.on_button_press(chr(key))
        elif key == Qt.Key_Period:
            self.on_button_press('.')
        elif key == Qt.Key_Plus:
            self.on_button_press('+')
        elif key == Qt.Key_Minus:
            self.on_button_press('-')
        elif key == Qt.Key_Asterisk:
            self.on_button_press('*')
        elif key == Qt.Key_Slash:
            self.on_button_press('/')
        elif key == Qt.Key_P:
            self.on_button_press('pow')


def main():
    app = QtWidgets.QApplication(sys.argv)
    ex = CalculatorApp()
    ex.show()
    sys.exit(app.exec_())
if __name__ == '__main__':
    main()
